/*
 *  UQLX
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/times.h>
#include <time.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>
#include <X11/Xlib.h> 
#include <X11/Xutil.h>
#include <X11/cursorfont.h>

#ifdef VM_SCR
#include <sys/mman.h>
#endif
  

#include "QLtypes.h"
#include "QL68000.h"
#include "unix.h"

#include "uqlx_cfg.h"

#define STATIC static

#ifdef SH_MEM
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>

static XShmSegmentInfo shminfo;
static shm_used;
#endif

XSetWindowAttributes attributes;

XWMHints wmhints;

int cursor_shape=XC_arrow;
extern int ki_state;
static int cs0=XC_arrow;
static int cs1=XC_cross;
Cursor cursor1,cursor0;
extern int invisible;
extern int start_iconic;


int screen_drawable=0;

int depth,inside;
long black,white;
extern int scr_width,scr_height;

extern void QLButton(int,int);
extern void QLMovePointer(int,int);

#if 0
extern char *scrModTable;
extern int sct_size;
extern int pagesize;
#endif

extern int verbose;

#define E_MASK ( KeyPressMask | KeyReleaseMask | ButtonPressMask  \
		 | ButtonReleaseMask | PointerMotionMask | ColormapChangeMask |\
		 EnterWindowMask | LeaveWindowMask | ExposureMask | ResizeRedirectMask |\
		 StructureNotifyMask )

     extern void QL2Pixmap8 (w32 , char *, w32 ,w32 );
     extern void QL2Pixmap16 (w32 , char *, w32 ,w32 );
     extern void QL2Pixmap24 (w32 , char *, w32 ,w32 );
     extern void QL2Pixmap32 (w32 , char *, w32 ,w32 );

     extern void (*QL2Pixmap) (w32 , char *, w32 ,w32 );
     

static int shmmono=1;
static char win_name[128];
static Visual *visual = NULL;

static int screen, nplanes;
static int Max_x, Max_y;

GC gc;
static XGCValues gc_val;

Window imagewin;		/* window ID for image */
XImage *image = NULL;
static char *Image;
Cursor cursor;
Pixmap empty;
int planes;
int pbytes;


char *xi_buf;

#ifdef SH_MEM
static int shmct;
#endif

Display *display;
int kmin,kmax;
int shmflag = 1;
Colormap cmap=0;

void redraw_screen(int ,int ,int ,int );
void ResetMousePtr(void);


void destroy_image (void)
{
  /*printf("Destroy Image\n");*/
  if (image == NULL) return;
  XDestroyImage (image);
#ifdef SH_MEM
  if (shmflag && shm_used)
    {
      XShmDetach (display, &shminfo);
      shmdt (shminfo.shmaddr);
      shmctl (shminfo.shmid, IPC_RMID, 0);
      shm_used = 0;
    }
#endif
  /*printf("End Destroy Image\n");*/
}

void x_reset_state(void)
{
#ifdef DO_GRAB
  if (ki_state==0)
    XGrabKey(display,AnyKey,AnyModifier,imagewin,False,GrabModeAsync,GrabModeAsync);
  else
    XUngrabKey(display,AnyKey,AnyModifier,imagewin);
#endif
  ResetMousePtr();
}

void x_screen_close (void)
{
  XAutoRepeatOn(display);
  XFlush(display);
  destroy_image();
  XCloseDisplay(display);
}

static void set_window_name (char *name)
{
  XStoreName (display, imagewin, name);
}




static void create_image (int w, int h, int format)
{
  int planes = depth;
  
#ifdef SH_MEM
  if (shmflag)
    {
      image = XShmCreateImage (display, visual, planes, format,
			       NULL, &shminfo, w, h);
      shminfo.shmid = shmget (IPC_PRIVATE, image->bytes_per_line
			      * image->height * ((planes+7)/8), IPC_CREAT | 0777);
      xi_buf = shminfo.shmaddr = image->data = shmat (shminfo.shmid, 0, 0);
      shminfo.readOnly = False;
      XShmAttach (display, &shminfo);
      XSync(display, False);
      shmctl(shminfo.shmid, IPC_RMID, 0);
      shm_used = 1;
#if 0
      bzero (&shminfo, sizeof(shminfo));
      printf("Using MIT Shm extension::\n\t\t If you get trouble with X errors\n");
      printf("\t\ttry to compile wthout the -DSH_MEM flag\n\n");
#endif

#ifdef BIG_ENDIAN
      image -> bitmap_bit_order = MSBFirst;
      image -> byte_order = MSBFirst;
#else
      image -> bitmap_bit_order = LSBFirst;
      image -> byte_order = LSBFirst;
#endif

      XShmPutImage(display,imagewin,gc,image,0,0,0,0,10,10,False);
      XSync(display,False);
    }

  /* XShmPutImage might have failed, setting shmflag=0 ! */

  if (!shmflag)
#endif
    {
      xi_buf= malloc(w*h*pbytes+256);
      
      image = XCreateImage (display, visual, planes, format,
			    0, (char *) xi_buf, w, h, 8, 0);
    }

#ifdef BIG_ENDIAN
  image -> bitmap_bit_order = MSBFirst;
  image -> byte_order = MSBFirst;
#else
  image -> bitmap_bit_order = LSBFirst;
  image -> byte_order = LSBFirst;
#endif
  /*Image = image->data;*/

}


/* Process all events that occured since we last came here...
 */

extern void xql_key(XEvent *,int);

void process_events (void)
{
  XEvent e;
  XExposeEvent *x;

  while (XCheckMaskEvent (display, E_MASK, &e))
    {
      switch (e.type)
	{
	case EnterNotify:
	  XAutoRepeatOff(display); /* QL has its autorepeat.. */
	  inside=1;
	  
	  break;
	case LeaveNotify:
	  XAutoRepeatOn(display);
	  inside=0;
	  
	  break;
	case ResizeRequest:
	  /* move way of making it not resizeable... */
	  XResizeWindow(display, imagewin, scr_width, scr_height);
	case KeyPress:
	  xql_key (&e, 1);
	  inside=1;
	  break;
	case KeyRelease:
	  xql_key (&e, 0);
	  break;
#ifdef MOUSE   /*mouse not yet supported */
	case ButtonPress:
	  QLButton(((XButtonEvent *)&e)->button,1);
	  inside=1;
	  break;
#endif
#ifdef MOUSE  /* no mouse yet */
	case ButtonRelease:
	  QLButton(((XButtonEvent *)&e)->button,0);
	  break;
#endif 

	case Expose:
	  /* flush all Expose events .. */
	  screen_drawable=1;
	  /* printf("***Expose event\n"); */
	  scrconv();
	  do{
#if 1
	    x = (XExposeEvent *)&e;
	    redraw_screen(x->x,x->y,x->width,x->height);
#endif
	  } while (XCheckTypedEvent(display,Expose,&e));
	  
	  XFlush(display);
	  break;

	case MapNotify:
	  screen_drawable=1;
	  /*printf("****MapNotify event\n");*/
	  break;
	  
	case UnmapNotify:
	  screen_drawable=0;
	  /*printf("****UnapNotify event\n");*/
	  break;
	  
#ifdef MOUSE 
	case MotionNotify: 
	  if (e.xmotion.window == imagewin)
	    {
	      QLMovePointer(((XMotionEvent *)&e)->x,((XMotionEvent *)&e)->y);
	      /*fprintf (stderr, "Pointer: %d,%d\n", ((XMotionEvent *)&e)->x,((XMotionEvent *)&e)->y); */
	      inside=1;
	    }
	  break;
#endif 0
	}
    }
}

long colors[16];

Visual * ChooseVisual(Display *display)
{
  XVisualInfo *visual_list;
  XVisualInfo vTemplate; 
  int vmatch,i,cmax;
  
  static char *visual_class[]={"StaticGray","GrayScale","StaticColor",
			       "PseudoColor","TrueColor","DirectColor"};

  int vpref_c[]={TrueColor,StaticColor,DirectColor,PseudoColor,-1};
  int vpref_m[]={StaticGray,GrayScale,-1};  
  int *p;
  

  vTemplate.screen=screen;
  vTemplate.depth=8;
  depth=vTemplate.depth;

  /*XSynchronize(display,1);*/

  visual_list=XGetVisualInfo(display, VisualScreenMask | VisualDepthMask,
			     &vTemplate, &vmatch);
  if (!vmatch)
    {
      vTemplate.depth=4;
      depth=vTemplate.depth;
      visual_list=XGetVisualInfo(display, VisualScreenMask | VisualDepthMask,
				 &vTemplate, &vmatch);
    }
  
  if (!vmatch)
    {
      printf("Warning: there seems to be no suitable visual\n");
      return XDefaultVisual(display,screen);
    }

  cmax=0;
  p=vpref_c;
  
  if (QMD.color)
    while (*p++ != -1)
      for (i=0;i<vmatch;i++)
	{
	  if (visual_list[i].colormap_size>8 && p[-1]==visual_list[i].class)
	    return visual_list[i].visual;
	}

  /* no color :-( */
  QMD.color=0;
  p=vpref_m;
  while ( *p++ != -1)
    for (i=0;i<vmatch;i++)
      {
	if (visual_list[i].colormap_size>8 && p[-1]==visual_list[i].class)
	  return visual_list[i].visual;
      }
  
  /* fallback into default visual .....*/
  depth = DefaultDepth (display, screen);
  return  XDefaultVisual (display, screen);
}


void allocColors(Colormap map)
{
#if 1
  static char *Cname[]={"black", "#0000FF","#FF0000","#FF00FF","#00FF00","#00FFFF","#FFFF00","#FFFFFF",  
              "#BFBFBF","#00BFFF","#FF00BF","#BF00FF","#BFFF00","#00FFBF","#FFBF00","#DFDFDF" };
  static char *Gname[]={"black","gray15","gray30","gray45","gray60","gray75","gray90","gray95",
              "gray5" ,"gray10","gray25","gray45","gray55","gray70","gray85","gray95"};
#else
  static char *Cname[]={"Black","Blue","Red","Magenta","Green","Cyan","Yellow","gray95" };
  static char *Gname[]={"Black","gray15","gray30","gray45","gray60","gray75","gray90","gray95"};
#endif
  char **name;
  
  XColor exact_def;
  /*int default_depth=DefaultDepth(display,screen); */
  int res,i=5;
  int i1=0;
  int lim;
  
  XVisualInfo visual_info;

  static char *visual_class[]={"StaticGray","GrayScale","StaticColor",
			       "PseudoColor","TrueColor","DirectColor"};

  planes=depth;
  /*plane2=planes/2;*/
  

  /*map=DefaultColormap(display,screen);*/
  /*XSetWindowColormap(display,imagewin,map);*/

  if (depth<4)  {
    printf("sorry, not much support fo monochrome display\n");
    colors[0]=colors[1]=black;
    colors[2]=colors[3]=white;
    colors[4]=colors[5]=black;
    colors[6]=colors[7]=white;
  }  

  /* do we have colour? */
  while(!XMatchVisualInfo(display,screen,depth,i--,&visual_info));
  
  if (verbose) printf("using Visual %s\n",visual_class[i++]);

  if (i>1 && QMD.color) name=Cname;
  else name=Gname;


  if (depth<8)
    {
      colors[0]=colors[1]=BlackPixel(display,screen);
      colors[7]=colors[6]=WhitePixel(display,screen);
      i1=1;
      lim=15/*7*/;
    }
  else
    {
      i1=0;
      lim=16/*8*/-QMD.fwhite;
    }
  
  
  for(i=i1; i<lim; i++)
    {
      res=XParseColor(display,map,name[i],&exact_def);
      /*printf("XParseColor returns: %d\n",res);*/
      res=XAllocColor(display,map,&exact_def);
      /*printf("XAllocColor returns: %d\n",res);*/
      colors[i]=exact_def.pixel;/*colors[2*(i+1)]=colors[2*(i+1)+1]=exact_def.pixel;*/
      /*printf("pixval : %d\n",exact_def.pixel);*/
    }

  if (QMD.fwhite)
    colors[7]=WhitePixel(display,screen);

  /*  for(i=0; i<8; i++)printf("Color %d= %d\n",i,colors[i]);*/
}

int xbreak=0;
extern int QLdone;
int myXIOerr(Display* d)
{
  printf("X connection closed\n");
  /*cleanup(0);*/
  QLdone=1;
  xbreak=1;
  cleanup(0);
}
int myXerr(Display* d, XErrorEvent* ev)
{
  if (shmflag) shmflag=0;
  else
    printf("X error : trying to ignore\n"); 
}

void setwmhints()
{
  wmhints.initial_state= start_iconic ? IconicState : NormalState;
  
  wmhints.flags=StateHint;
}


void x_screen_open (int frk)
{
  unsigned long valuemask;
  int i;
  char *ip;
  
#ifdef SH_MEM
  int shm_major, shm_minor;
  Bool shm_pixmaps;
#endif

  XSizeHints h;
  XColor ccls[2];


  scr_width = qlscreen.xres;
  scr_height = qlscreen.yres;


  /*strcpy (win_name, "QL");*/
  sprintf(win_name,"QL - %s, %dK",QMD.sysrom,RTOP/1024);

  if ((display = XOpenDisplay (NULL)) == NULL)
    {
      perror(NULL);
      fprintf (stderr, "Could not open display.\n");
      exit (1);
    }
 
  if (!frk)
    init_keymap();
  
  /* set X error handler */
  XSetErrorHandler(myXerr);
  XSetIOErrorHandler(myXIOerr);
#ifdef SH_MEM 
  if (shmflag)
    {
      if (!XShmQueryVersion (display, &shm_major, &shm_minor, &shm_pixmaps))
	{
	  verbose && fprintf(stderr,
			     "MIT Shared Memory Extension not supported.\n");
	  shmflag = 0;
	}
      else
	{
	  int foo;
	  verbose && fprintf (stderr,
			      "Using MIT Shared Memory Extension %d.%d," \
			      " %s shared pixmaps.\n", shm_major, shm_minor,
			      (shm_pixmaps ? "with" : "without"));
	  foo = XShmPixmapFormat (display);
#if 0
	  verbose && fprintf (stderr,
			      "XShm Pixmap format: %s\n", ((foo == XYBitmap ? 
							    "XYBitmap" : (foo == XYPixmap ? "XYPixmap" : "ZPixmap"))));
#endif
	}
    }
#endif

  screen = XDefaultScreen (display);

#if XVIS	
  visual=ChooseVisual(display);
  nplanes=depth;
#else

  depth = DefaultDepth (display, screen);
  visual = XDefaultVisual (display, screen);
  nplanes = XDisplayPlanes (display, screen);
#endif

  if (nplanes<=16) pbytes=(nplanes+7)/8;
  else pbytes=4;
  

  h.width = scr_width;
  h.height = scr_height; 
  h.x = 500;
  h.y = 0;

  h.flags = PSize; /*| PPosition; /*  ?????? */
  black=BlackPixel (display, screen);
  white = WhitePixel (display, screen);
  
  attributes.backing_store=/*start_iconic ? WhenMapped : */ Always;
  attributes.border_pixel=BlackPixel(display,screen);
  attributes.background_pixel=BlackPixel(display,screen);
  valuemask= CWBackingStore | CWBorderPixel | CWBackPixel |CWColormap; 
 
  if (visual==XDefaultVisual(display, screen) && depth>4)
    cmap=DefaultColormap(display,screen);
  else
    cmap=XCreateColormap(display, RootWindow (display, screen), visual,AllocNone);

  attributes.colormap=cmap;
  
  imagewin = XCreateWindow (display, RootWindow (display, screen),
			    h.x, h.y, h.width, h.height, 0, depth,
			    InputOutput, visual, valuemask, &attributes);
  XSetWindowColormap(display,imagewin,cmap);
  
  setwmhints();
  XSetWMHints(display, imagewin, &wmhints);
  
  XSelectInput (display, imagewin, StructureNotifyMask|E_MASK);
  XSetStandardProperties (display, imagewin, win_name, win_name, None,
			  NULL, 0, &h);


  XMapWindow (display, imagewin);
  if (!start_iconic)
    {
      for (;;)
	{
	  XEvent e;
	  XNextEvent (display, &e);
	  if (e.type == MapNotify && e.xmap.event == imagewin) break;
	}
      screen_drawable=1;
    }
  else screen_drawable=0;
  
  gc_val.background = black;
  gc_val.foreground = white;
  gc = XCreateGC (display, imagewin, GCForeground | GCBackground, &gc_val);
#ifdef SH_MEM
  shmct = XShmGetEventBase (display) + ShmCompletion;
#endif
  XSetForeground (display, gc, white);
  XSetBackground (display, gc, black);

  if(!frk)
    allocColors(cmap);

  XClearWindow (display, imagewin);

  XSelectInput (display, imagewin, E_MASK);

  if (1 /*!frk*/)
    create_image (scr_width,scr_height,ZPixmap);


  if (nplanes<=8) QL2Pixmap=QL2Pixmap8;
  
  if (8 < nplanes && nplanes <= 16) QL2Pixmap=QL2Pixmap16;

  if (16 < nplanes && nplanes <= 24) QL2Pixmap=QL2Pixmap24;
  
  if (24 < nplanes && nplanes <= 32) QL2Pixmap=QL2Pixmap32;

  if (nplanes>32)
    {
      printf("sorry, unsupported screen depth %d\n",nplanes);
      QL2Pixmap=QL2Pixmap32;
    }


  /* define XCursors used */
  ccls[0].pixel=0;
  ccls[1].pixel=0;
  empty =  XCreatePixmapFromBitmapData (display, imagewin, calloc(32,1),
					16, 16,0,0,1);
 
  cursor0= XCreateFontCursor(display,cs0);
  cursor1= XCreateFontCursor(display,cs1);
  cursor = XCreatePixmapCursor (display,empty,empty, &ccls[0],&ccls[1], 0,0);

  /* Set Passive Grabs for every key that could be snatched away by the WM */

  x_reset_state();
  
#if 0 /* done by x_reset_state */
  XGrabKey(display,AnyKey,AnyModifier,imagewin,False,GrabModeAsync,GrabModeAsync);
#endif

}



void ResetMousePtr()
{
#if 0
  XUndefineCursor(display,imagewin);
#else
  if (invisible==0)
    {
      if (ki_state)
	XDefineCursor(display,imagewin,cursor1);
      else
	XDefineCursor(display,imagewin,cursor0);
    }
  else InvisibleMousePtr();
#endif
}
void InvisibleMousePtr()
{
  XDefineCursor(display,imagewin,cursor);
}


